"""Perform the main functionality of the application.

Copyright (C) 2021 emerac

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
import logging
import sys
from typing import List, Optional

from filmbuff import arguments, interface


logger = logging.getLogger(__name__)


def main(argv: Optional[List[str]] = None) -> None:
    """Handle command-line arguments and run the application.

    Parameters
    ----------
    argv
        Arguments to pass directly to the parser and QApplication.

    Returns
    -------
    None

    """
    args = arguments.parse_args(argv)
    arguments.configure_logging(args.verbose)

    if args.create_shortcut:
        arguments.create_shortcut()
        return

    logger.info("Starting filmbuff.")
    interface.main(argv)
    logger.info("Exiting filmbuff.")


if __name__ == "__main__":
    main(sys.argv[1:])



"""Prepare SQLite WHERE clauses for use in model filtering.

Copyright (C) 2021 emerac

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
import re
from typing import Dict, List, Pattern


def add_quotes(text: str) -> str:
    """Add double quotes to the start and end of a string.

    Double quotes will not be added to indexes that already contain
    them.

    Parameters
    ----------
    text
        The string to wrap in double quotes.

    Returns
    -------
    str
        The string wrapped in double quotes.

    """
    if not text.startswith("\""):
        text = "\"" + text
    if not text.endswith("\""):
        text = text + "\""
    return text


def escape_query(query: str) -> str:
    """Escape special SQLite characters appearing in the query.

    A backslash will be added before each instance of '%' and '_'
    appearing in the query.

    Parameters
    ----------
    query
        The query potentially containing special characters.

    Returns
    -------
    str
        The query with all instances of '%' and '_' escaped.

    Notes
    -----
    When the final WHERE clause is assembled, any characters preceded
    by the designated ESCAPE character (a backslash), will be treated
    literally.

    See Also
    --------
    handle_regular_query :
        Create a valid SQLite WHERE clause based on the regular query.
    handle_special_query :
        Create a valid SQLite WHERE clause based on the special query.

    """
    return query.replace("%", "\\%").replace("_", "\\_")


def expand_query(query: str) -> str:
    """Convert an unexpanded query into its expanded equivalent.

    The expanded query uses the same search syntax as if each column
    had been explicitly included in the search.

    Parameters
    ----------
    query
        The query containing one or more expansion abbreviations.

    Returns
    -------
    str
        The expanded query.

    Warnings
    --------
    It is not recommended to use expanding and non-expanding queries
    in the same search. When a query is expanded, no parentheses are
    placed around the expansion, which can lead to unexpected results.
    See Example 2.

    Examples
    --------
    Example 1: Standard usage.

    >>> expand_query('"genre":"x"')
    '"genre one":"x" or "genre two":"x" or "genre three":"x"'

    Example 2: Expanding and non-expanding queries should not be used
    in the same query.

    >>> expand_query('"cast":"x" and "runtime":"y"')
    '"cast one":"x" or "cast two":"x" or "cast three":"x" and
    "runtime":"y"'

    The result above is intended behavior. However, one might expect
    the search to operate in the following manner (notice the
    parentheses):

    ("cast one":"x" or "cast two":"x" or "cast three":"x") and
    "runtime":"y"

    See Also
    --------
    is_expanding_query : Determine if the query should be expanded.

    """
    subqueries = get_subqueries(query, True)
    cast_regex = re.compile(r"\"cast\":\"(.*)\"", re.IGNORECASE)
    genre_regex = re.compile(r"\"genre\":\"(.*)\"", re.IGNORECASE)
    operator_regex = re.compile(r"^\"?(and|or)\"?$", re.IGNORECASE)
    for index, subquery in enumerate(subqueries):
        cast_query_mo = cast_regex.search(subquery)
        if cast_query_mo is not None:
            expanded_query = prepare_expansion("cast", cast_regex, subquery)
            subqueries[index] = expanded_query

        genre_query_mo = genre_regex.search(subquery)
        if genre_query_mo is not None:
            expanded_query = prepare_expansion("genre", genre_regex, subquery)
            subqueries[index] = expanded_query

        operator_mo = operator_regex.search(subquery)
        if operator_mo is not None:
            formatted_operator = remove_quotes(operator_mo.group(0))
            subqueries[index] = formatted_operator
    return " ".join(subqueries)


def format_header_name(header_name: str) -> str:
    """Convert a raw header to the correct display format.

    Parameters
    ----------
    header_name
        The header name to format.

    Returns
    -------
    str
        The header name formatted to exactly match a display header
        name.

    """
    if "id" in header_name:
        header_name = header_name.upper()
    else:
        header_name = header_name.capitalize()
    return header_name


def get_special_operators(query: str) -> List[str]:
    """Retrieve instances of 'and' or 'or' in special queries.

    Parameters
    ----------
    query
        A query containing special query syntax.

    Returns
    -------
    list of str
        The operators in order of appearance from left to right.

    """
    operators_regex = re.compile(r"\"\s(and|or)\s\"", re.IGNORECASE)
    operators = operators_regex.findall(query)
    return operators


def get_subqueries(query: str, operators: bool) -> List[str]:
    """Split a query into subqueries, optionally including operators.

    Parameters
    ----------
    query
        A query containing special query syntax.
    operators
        Whether or not to include operators in the list of subqueries.

    Returns
    -------
    list of str
        The individual subqueries that make up `query`.

    """
    subquery_regex = re.compile(r"\"\s(and|or)\s\"", re.IGNORECASE)
    subqueries = subquery_regex.split(query)
    temps = []
    for subquery in subqueries:
        temp = subquery
        temp = add_quotes(temp)
        if temp.lower() == "\"and\"" or temp.lower() == "\"or\"":
            if not operators:
                continue
        temps.append(temp.strip())
    return temps


def handle_regular_query(query: str, headers: Dict[str, str]) -> str:
    """Create a valid SQLite WHERE clause based on a regular query.

    Parameters
    ----------
    query
        A query that does not contain special query syntax.
    headers
        Header names, as displayed in QTableView, mapped to database
        column names.

    Returns
    -------
    str
        A SQLite WHERE clause that searches across all database
        columns. Intended for use with QSqlTableModel.setFilter.

    """
    formatted_query = escape_query(query.strip())
    sql_substrings = [
        f"{column} LIKE \"%{formatted_query}%\" ESCAPE \'\\\' OR "
        for column in headers.values()
    ]
    sql_string = "".join(sql_substrings)[:-4]
    return sql_string


def handle_special_query(query: str, headers: Dict[str, str]) -> str:
    """Create a valid SQLite WHERE clause based on a special query.

    Parameters
    ----------
    query
        A query that contains valid special query syntax.
    headers
        Header names, as displayed in QTableView, mapped to database
        column names.

    Returns
    -------
    str
        A SQLite WHERE clause that searches across only specified
        database columns. Intended for use with
        QSqlTableModel.setFilter.

    """
    formatted_query = escape_query(query.strip())
    subqueries = get_subqueries(formatted_query, False)
    operators = get_special_operators(formatted_query)

    if not is_valid_special_query(subqueries, operators, headers):
        raise ValueError("Expression is not a valid special query.")

    query_regex = re.compile(r"\"(.*)\":\"(.*)\"", re.IGNORECASE)
    noop_sql_substrings = []
    for subquery in subqueries:
        query_mo = query_regex.search(subquery)
        assert query_mo is not None

        header_name = query_mo.group(1).lower()
        formatted_header_name = format_header_name(header_name)
        column = headers[formatted_header_name]
        term = query_mo.group(2)

        noop_sql_substrings.append(
            f"{column} LIKE \"%{term}%\" ESCAPE \'\\\' OPERATOR "
        )

    operator_regex = re.compile(r"OPERATOR")
    sql_substrings = []
    for index, substring in enumerate(noop_sql_substrings):
        if index == len(operators):
            sql_substrings.append(substring)
        else:
            sub = operator_regex.sub(operators[index].upper(), substring)
            sql_substrings.append(sub)
    sql_string = "".join(sql_substrings)[:-10]
    return sql_string


def is_expanding_query(query: str) -> bool:
    """Determine if a query should be expanded.

    Expanding queries are denoted by the syntax '"cast":"x"' or
    '"genre":"x"'.

    Parameters
    ----------
    query
        A raw query.

    Returns
    -------
    bool
        True if the query should be expanded, False if not.

    Notes
    -----
    The database contains three cast columns and three genre columns.
    To avoid having to create a tediously long query to search across
    a category's three columns, two abbreviations ('cast' and 'genre')
    can be used instead. The existence of such abbreviations, along
    with the correct syntax, marks the query as needing to expanded
    into its equivalent long form.

    """
    if "\"cast\":" in query or "\"genre\":" in query:
        return True
    else:
        return False


def is_special_query(query: str, headers: Dict[str, str]) -> bool:
    """Determine if the query contains any special query syntax.

    Special queries are denoted by the syntax: '"x":', where 'x'
    is a header name. Regular queries are those that are not special
    queries.

    Parameters
    ----------
    query
        A query potentially containing special query syntax.
    headers
        Header names, as displayed in QTableView, mapped to database
        column names.

    Returns
    -------
    bool
        True if the query contains any special query syntax, False if
        not.

    Notes
    -----
    This function does not check whether or not any special query
    syntax found is valid.

    See Also
    --------
    is_valid_special_query : Determine if the special query is valid.

    """
    for header in headers.keys():
        if (
            f"\"{header.lower()}\":" in query.lower()
            or "\"cast\":" in query.lower()
            or "\"genre\":" in query.lower()
        ):
            return True
    return False


def is_valid_special_query(
    subqueries: List[str],
    operators: List[str],
    headers: Dict[str, str],
) -> bool:
    """Determine if the special query is valid.

    To be considered valid, a special query must be entirely correct.

    Parameters
    ----------
    subqueries
        The query split on operators.
    operators
        The operators on which the query was split.
    headers
        Header names, as displayed in QTableView, mapped to database
        column names.

    Returns
    -------
    bool
        True if the entire special query is valid, False if not.

    See Also
    --------
    is_special_query :
        Determine if the query contains any special query syntax.

    """
    all_headers = [header.lower() for header in headers.keys()]
    all_headers.append("cast")
    all_headers.append("genre")

    subquery_regex = re.compile(r"\"(.*)\":\"(.*)\"", re.IGNORECASE)
    for subquery in subqueries:
        subquery_mo = subquery_regex.search(subquery)
        if subquery_mo is None:
            return False

        user_header = subquery_mo.group(1).lower()
        if user_header not in all_headers:
            return False

    for operator in operators:
        if operator.lower() not in ["and", "or"]:
            return False
    return True


def prepare_expansion(abbrev: str, regex: Pattern[str], text: str) -> str:
    """Convert unexpanded text into its expanded form.

    Parameters
    ----------
    abbrev: {'cast', 'genre'}
        The display name to expand.
    regex
        The regex used to locate search terms.
    text
        The text to expand.

    Returns
    -------
    str
        The text in its expanded form, suitable for replacement.

    """
    text_mo = regex.search(text)
    assert text_mo is not None
    expansion = (
        f"\"{abbrev} one\":\"{text_mo.group(1)}\" "
        f"or \"{abbrev} two\":\"{text_mo.group(1)}\" "
        f"or \"{abbrev} three\":\"{text_mo.group(1)}\""
    )
    expanded_query = regex.sub(text, expansion)
    return expanded_query


def remove_quotes(text: str) -> str:
    """Remove double quotes from the start and end of a string.

    Quotes will not be removed if the result would be an empty string.

    Parameters
    ----------
    text
        The string from which to remove double quotes.

    Returns
    -------
    str
        The string with no quotes at the beginning or ending indices.

    """
    if text == "\"\"":
        return text
    if text.startswith("\""):
        text = text[1:]
    if text.endswith("\""):
        text = text[:-1]
    return text


def main(query: str, headers: Dict[str, str]) -> str:
    """Create a valid SQLite WHERE clause from a raw query.

    This function will handle a raw query, start-to-finish, performing
    any necessary expansions and formatting.

    Parameters
    ----------
    query
        The raw query.
    headers
        Header names, as displayed in QTableView, mapped to database
        column names.

    Returns
    -------
    str
        A SQLite WHERE clause used to search across database columns.
        Intended for use with QSqlTableModel.setFilter.

    """
    if is_expanding_query(query):
        query = expand_query(query)

    if is_special_query(query, headers):
        sql_string = handle_special_query(query, headers)
    else:
        sql_string = handle_regular_query(query, headers)
    return sql_string



"""Parse the command-line arguments and configure logging.

Copyright (C) 2021 emerac

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
import argparse
from importlib.resources import path
import logging
from pathlib import Path
import shutil
from typing import List, Optional

import filmbuff


def create_shortcut() -> None:
    """Create a shortcut to the filmbuff application.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Notes
    -----
    This function copies a prepared desktop entry file into an XDG
    Desktop Entry directory.

    If filmbuff is installed in a virtual env or not at all, the exec
    key in the .desktop file will be invalid and the entry will not show
    up.

    """
    with path("filmbuff.resources", "filmbuff.desktop") as resource_path:
        entry_src = resource_path
    entry_dest = (
        Path.home()
        / ".local"
        / "share"
        / "applications"
        / "filmbuff.desktop"
    )
    shutil.copy(entry_src, entry_dest)


def configure_logging(verbose: int) -> None:
    """Set the log handler and level.

    Parameters
    ----------
    verbose
        The level of verbosity desired.

    Returns
    -------
    None

    """
    if verbose == 0:
        filmbuff.logger.addHandler(filmbuff.null_handler)
    else:
        filmbuff.logger.addHandler(filmbuff.stream_handler)
        if verbose == 1:
            filmbuff.logger.setLevel(logging.INFO)
        else:
            filmbuff.logger.setLevel(logging.DEBUG)


def parse_args(argv: Optional[List[str]] = None) -> argparse.Namespace:
    """Create a parser and parse command-line arguments.

    If no arguments are provided, the parser reads from `sys.argv`.

    Parameters
    ----------
    argv
        Arguments to pass directly to the parser.

    Returns
    -------
    :obj:`argparse.Namespace`
        A named-tuple like object containing the parsed arguments.

    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=(
            "filmbuff is an application for creating and managing a local\n"
            "database of film information."
            "\n\n"
            "This product uses the TMDb API but is not endorsed or certified\n"
            "by TMDb."
        ),
        epilog=(
            "filmbuff Copyright (C) 2021 emerac"
            "\n\n"
            "This program comes with ABSOLUTELY NO WARRANTY. This is free\n"
            "software, and you are welcome to redistribute it under certain\n"
            "conditions. You should have received a copy of the GNU General\n"
            "Public License along with this program. If not, see:\n"
            "<https://www.gnu.org/licenses/gpl-3.0.txt>."
        ),
    )
    parser.add_argument(
        "--verbose",
        "-v",
        help="Provide log output. This argument can be repeated.",
        action="count",
        default=0,
    )
    parser.add_argument(
        "--create-shortcut",
        help=(
            "Create a shortcut in the applications menu. To be used on GNOME "
            "desktop environments only. This shortcut will not work if "
            "filmbuff is installed in a virtual environment."
        ),
        action="store_true",
    )
    return parser.parse_args(argv)



"""Provide 'about' and 'help' windows.

Copyright (C) 2021 emerac

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
from importlib.resources import path

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QPalette, QShowEvent
from PyQt5.QtSvg import QSvgWidget
from PyQt5.QtWidgets import (
    QHBoxLayout,
    QLabel,
    QMainWindow,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)


class AboutWindow(QWidget):
    """A window used to present application information.

    Notes
    -----
    Methods whose names are written in mixedCase are overridden
    QWidget methods.

    """

    def __init__(self, main_win: QMainWindow) -> None:
        """Construct an AboutWindow instance.

        Parameters
        ----------
        main_win
            The AboutWindow's parent window.

        Returns
        -------
        None

        """
        super().__init__()
        self.main_win = main_win
        self.setWindowTitle("About - filmbuff")

        self.title = QLabel()
        self.title.setText("filmbuff")
        self.title.setAlignment(Qt.AlignCenter)
        self.title.setStyleSheet(
            "font-weight: bold;"
            "font-size: 26px;"
            "padding-top: 0;"
            "padding-bottom: 4px;"
        )

        self.description = QLabel()
        self.description.setText(
            "Create and manage local databases of film information."
        )
        self.description.setAlignment(Qt.AlignCenter)
        self.description.setStyleSheet(
            "padding: 4px;"
        )

        self.copyright_info = QLabel()
        self.copyright_info.setText(
            "filmbuff Copyright © 2021 emerac\n"
            "<https://gitlab.com/emerac/filmbuff>"
        )
        self.copyright_info.setAlignment(Qt.AlignCenter)
        self.copyright_info.setStyleSheet(
            "font-size: 12px;"
            "padding: 4px;"
        )

        self.disclaimer = QLabel()
        self.disclaimer.setText(
            "This program comes with ABSOLUTELY NO WARRANTY. This is free\n"
            "software, and you are welcome to redistribute it under certain\n"
            "conditions. You should have received a copy of the GNU General\n"
            "Public License along with this program. If not, see:\n"
            "<https://www.gnu.org/licenses/gpl-3.0.txt>."
        )
        self.disclaimer.setAlignment(Qt.AlignCenter)
        self.disclaimer.setStyleSheet(
            "font-size: 12px;"
            "padding-top: 4px;"
            "padding-bottom: 8px;"
        )

        self.logo_container = QHBoxLayout()
        self.logo_container.setAlignment(Qt.AlignCenter)
        self.tmdb_logo = QSvgWidget()
        with path("filmbuff.resources", "tmdb_logo.svg") as logo_path:
            self.tmdb_logo.load(str(logo_path))
        # Scaled to ~40% of original dimensions.
        self.tmdb_logo.setFixedSize(109, 14)
        self.logo_container.addWidget(self.tmdb_logo)

        self.attribution = QLabel()
        self.attribution.setText(
            "This product uses the TMDb API but is not endorsed or certified\n"
            "by TMDb."
        )
        self.attribution.setAlignment(Qt.AlignCenter)
        self.attribution.setStyleSheet(
            "font-size: 12px;"
            "padding-top: 4px;"
            "padding-bottom: 0;"
        )

        main_layout = QVBoxLayout()
        main_layout.addWidget(self.title)
        main_layout.addWidget(self.description)
        main_layout.addWidget(self.copyright_info)
        main_layout.addWidget(self.disclaimer)
        main_layout.addLayout(self.logo_container)
        main_layout.addWidget(self.attribution)
        self.setLayout(main_layout)

    def showEvent(self, event: QShowEvent) -> None:
        """Handle QShowEvents.

        This method waits for QShowEvents to occur. When a QShowEvent is
        received, the window will be made to be centered over the parent
        window.

        Parameters
        ----------
        event
            Information about the show event that occurred.

        Returns
        -------
        None

        Notes
        -----
        This is an overridden baseclass (QWidget) method.

        This method handles QShowEvents, which are automatically
        produced in response to widget show signals. When a QShowEvent
        occurs, the QShowEvent is automatically forwarded to this
        method. QShowEvents can be manufactured and passed to this
        method as well.

        """
        super().showEvent(event)
        self.center_window()
        self.setFixedSize(self.width(), self.height())

    def center_window(self) -> None:
        """Center the window over its parent window.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        frame_width = self.main_win.frameGeometry().width()
        win_width = self.main_win.width()
        header_width = frame_width - win_width

        frame_height = self.main_win.frameGeometry().height()
        win_height = self.main_win.height()
        header_height = frame_height - win_height

        center_x = self.main_win.x() + header_width + 0.5 * win_width
        center_y = self.main_win.y() + header_height + 0.5 * win_height
        win_x = center_x - 0.5 * self.width()
        win_y = center_y - 0.5 * self.height()
        self.move(int(win_x), int(win_y))


class HelpWindow(QWidget):
    """A window used to present application help.

    Notes
    -----
    Methods whose names are written in mixedCase are overridden
    QWidget methods.

    """

    def __init__(self, main_win: QMainWindow) -> None:
        """Construct a HelpWindow instance.

        Parameters
        ----------
        main_win
            The HelpWindow's parent window.

        Returns
        -------
        None

        """
        super().__init__()
        self.main_win = main_win
        self.setWindowTitle("Help - filmbuff")

        with path("filmbuff.resources", "help.md") as text_path:
            with open(text_path) as f:
                self.text = f.read()
        self.display = QTextEdit(self.text)
        self.display.setMarkdown(self.text)
        self.display.setReadOnly(True)

        main_win_palette = self.main_win.palette()
        background_color = main_win_palette.color(
            QPalette.Normal,
            QPalette.Window,
        )
        display_palette = self.display.palette()
        display_palette.setColor(
            QPalette.Normal,
            QPalette.Base,
            background_color,
        )
        display_palette.setColor(
            QPalette.Inactive,
            QPalette.Base,
            background_color,
        )
        self.display.setPalette(display_palette)
        self.display.setViewportMargins(10, 5, 10, 5)

        self.main_layout = QHBoxLayout()
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.addWidget(self.display)
        self.setLayout(self.main_layout)

    def showEvent(self, event: QShowEvent) -> None:
        """Handle QShowEvents.

        This method waits for QShowEvents to occur. When a QShowEvent is
        received, the window will be made to be centered over the parent
        window.

        Parameters
        ----------
        event
            Information about the show event that occurred.

        Returns
        -------
        None

        Notes
        -----
        This is an overridden baseclass (QWidget) method.

        This method handles QShowEvents, which are automatically
        produced in response to widget show signals. When a QShowEvent
        occurs, the QShowEvent is automatically forwarded to this
        method. QShowEvents can be manufactured and passed to this
        method as well.

        """
        super().showEvent(event)
        self.resize_window()
        self.center_window()

    def center_window(self) -> None:
        """Center the window over its parent window.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        frame_width = self.main_win.frameGeometry().width()
        win_width = self.main_win.width()
        header_width = frame_width - win_width

        frame_height = self.main_win.frameGeometry().height()
        win_height = self.main_win.height()
        header_height = frame_height - win_height

        center_x = self.main_win.x() + header_width + 0.5 * win_width
        center_y = self.main_win.y() + header_height + 0.5 * win_height
        win_x = center_x - 0.5 * self.width()
        win_y = center_y - 0.5 * self.height()
        self.move(int(win_x), int(win_y))

    def resize_window(self) -> None:
        self.resize(500, 600)



"""Provide a graphic user interface.

Copyright (C) 2021 emerac

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
import csv
import logging
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, cast

from PyQt5.QtCore import (
    QCoreApplication,
    QModelIndex,
    QSettings,
    QStandardPaths,
    Qt,
)
from PyQt5.QtGui import (
    QCloseEvent,
    QCursor,
    QKeyEvent,
    QIcon,
    QKeySequence,
    QResizeEvent,
    QShowEvent,
)
from PyQt5.QtSql import (
    QSqlDatabase,
    QSqlQuery,
    QSqlRecord,
    QSqlTableModel,
)
from PyQt5.QtWidgets import (
    QAction,
    QApplication,
    QCheckBox,
    QDialog,
    QFileDialog,
    QFormLayout,
    QHeaderView,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMainWindow,
    QMenu,
    QMessageBox,
    QPushButton,
    QSizePolicy,
    QSpinBox,
    QStatusBar,
    QTableView,
    QToolBar,
    QVBoxLayout,
    QWidget,
)
import send2trash  # type: ignore  # no stubs provided

from filmbuff import information, metadata, tablequery


logger = logging.getLogger(__name__)


class FileDialog(QFileDialog):
    """A configurable QFileDialog for user-filesystem interaction.

    Intended use cases include: asking the user which file to open and
    asking the user under which name and location they would like to
    save a file.

    """

    def __init__(self, **kwargs: Any) -> None:
        """Construct a FileDialog instance.

        Parameters
        ----------
        **kwargs
            A dictionary of FileDialog options.
            Each key corresponds to a QFileDialog method with that key's
            value being directly forwarded to the corresponding method.
            The key's and their corresponding methods are listed below:
                parent - __init__
                path - setDirectory
                file_mode - setFileMode
                accept_mode - setAcceptMode
                name_filter - setNameFilter
                options - setOptions

        Returns
        -------
        None

        """
        super().__init__(kwargs.get("parent"))
        self.setWindowModality(Qt.WindowModal)
        if kwargs.get("path") is not None:
            self.setDirectory(kwargs["path"])
        if kwargs.get("file_mode") is not None:
            self.setFileMode(kwargs["file_mode"])
        if kwargs.get("accept_mode") is not None:
            self.setAcceptMode(kwargs["accept_mode"])
        if kwargs.get("name_filter") is not None:
            self.setNameFilter(kwargs["name_filter"])
        if kwargs.get("options") is not None:
            self.setOptions(kwargs["options"])


class Interface(QMainWindow):
    """Primary point of interaction between the user and application.

    Notes
    -----
    Methods whose names are written in mixedCase are overridden
    QMainWindow methods.

    """

    def __init__(self) -> None:
        """Construct an Interface instance.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        super().__init__()
        self.setWindowTitle("filmbuff")
        self.setWindowIcon(QIcon.fromTheme("x-office-spreadsheet"))

        QCoreApplication.setOrganizationName("filmbuff")
        QCoreApplication.setApplicationName("filmbuff")
        self.settings = QSettings()
        self.data_path = self.make_data_path()

        recent_files = self.settings.value("recent_files")
        if recent_files is None or not recent_files[-1].exists():
            self.con_name = "virtual_1"
            open_db(self.con_name, ":memory:")
            create_empty_table(self.con_name)
            self.set_window_title()
        else:
            self.con_name = "physical_1"
            open_db(self.con_name, str(recent_files[-1]))
            self.set_window_title()

        self._configure_model()
        self._configure_view()
        self._configure_header()
        self._configure_actions()
        self._configure_menubar()
        self._configure_toolbar()
        self._configure_header_context()

        self.statusbar = StatusBar(5000, self)
        self.setStatusBar(self.statusbar)

        self.update_recent_files_menu()

        self.setCentralWidget(self.view)

        self.resize_and_center()

        self.model.select()

        self.about_window: Optional[information.AboutWindow] = None
        self.help_window: Optional[information.HelpWindow] = None
        self.metadata_window: Optional[metadata.MetadataWindow] = None

    def closeEvent(self, event: QCloseEvent) -> None:
        """Handle QCloseEvents.

        This method waits for QCloseEvents to occur. Given a
        QCloseEvent, it will ensure that the state of the application
        (window geometry and header configuration) are saved and that
        the application closes. The application will not close if the
        model contains unsaved changes.

        Parameters
        ----------
        event
            Information about the close event that occurred.

        Returns
        -------
        None

        Notes
        -----
        This is an overridden baseclass (QMainWindow) method.

        This method handles QCloseEvents, which are automatically
        produced in response to window close signals. When a QCloseEvent
        occurs, the QCloseEvent is automatically forwarded to this
        method. QCloseEvents can be manufactured and passed to this
        method as well.

        """
        super().closeEvent(event)
        self.save_window_geometry()
        self.save_header_state()

        if not self.is_state_resolved():
            event.ignore()

    def keyPressEvent(self, event: QKeyEvent) -> None:
        """Handle QKeyEvents.

        This method waits for certain QKeyEvents to occur. Given the
        appropriate QKeyEvent, it will either focus on the searchbar
        or cycle through the records displayed by the view.

        Parameters
        ----------
        event
            Information about the key event that occurred.

        Returns
        -------
        None

        Notes
        -----
        This is an overridden baseclass (QMainWindow) method.

        This method handles QKeyEvents, which are automatically produced
        in response to key presses. When a QKeyEvent occurs, the
        QKeyEvent is automatically forwarded to this method. QKeyEvents
        can be manufactured and passed to this method as well.

        """
        mod = cast(Qt.KeyboardModifier, event.modifiers())
        if mod == Qt.ControlModifier and event.key() == Qt.Key_F:
            self.searchbar.setFocus()

        if self.view.hasFocus() and event.key() == Qt.Key_Return:
            # Find the first selectable and non-hidden column.
            current_row = self.view.currentIndex().row()
            max_row = self.model.rowCount() - 1
            first_visible = 0
            for visual_index in range(self.header.count()):
                logical_index = self.header.logicalIndex(visual_index)
                if self.model.headerData(logical_index, Qt.Horizontal) == "ID":
                    continue
                if not self.header.isSectionHidden(logical_index):
                    first_visible = self.header.logicalIndex(visual_index)
                    break

            if current_row == -1:
                return
            elif current_row == max_row:
                new_index = self.model.createIndex(0, first_visible)
                self.view.setCurrentIndex(new_index)
            else:
                new_index = self.model.createIndex(
                    current_row + 1,
                    first_visible,
                )
                self.view.setCurrentIndex(new_index)
            current_row = self.view.currentIndex().row()
            assert 0 <= current_row <= max_row

    def resizeEvent(self, event: QResizeEvent) -> None:
        """Handle QResizeEvents.

        This method waits for certain QResizeEvents to occur. Given the
        appropriate QResizeEvent, it will ensure that the searchbar is
        viewable even when the window width is small.

        Parameters
        ----------
        event
            Information about the resize event that occurred.

        Returns
        -------
        None

        Notes
        -----
        This is an overridden baseclass (QMainWindow) method.

        This method handles QResizeEvents, which are automatically
        produced in response to window resizing. When a QResizeEvent
        occurs, the QResizeEvent is automatically forwarded to this
        method. QResizeEvents can be manufactured and passed to this
        method as well.

        """
        super().resizeEvent(event)
        if self.width() > 900:
            self.spacer.setSizePolicy(
                QSizePolicy.Expanding,
                QSizePolicy.Preferred,
            )
            self.searchbar.setMinimumWidth(570)
        else:
            self.spacer.setSizePolicy(
                QSizePolicy.Minimum,
                QSizePolicy.Preferred,
            )
            self.searchbar.setMinimumWidth(0)

    def showEvent(self, event: QShowEvent) -> None:
        """Handle QShowEvents.

        This method waits for QShowEvents to occur. Given the
        appropriate QShowEvent, it will reload the previous application
        state.

        Parameters
        ----------
        event
            Information about the show event that occurred.

        Returns
        -------
        None

        Notes
        -----
        This is an overridden baseclass (QMainWindow) method.

        This method handles QShowEvents, which are automatically
        produced in response to window show signals. When a QShowEvent
        occurs, the QShowEvent is automatically forwarded to this
        method. QShowEvents can be manufactured and passed to this
        method as well.

        """
        super().showEvent(event)
        self.save_default_window_geometry()
        self.save_default_header_state()
        self.load_settings()
        self.set_header_check_states()

        status = f"All records displayed ({self.model.rowCount()})."
        self.statusbar.showMessage(status, self.statusbar.timeout)

    def data_export(self) -> None:
        """Export the current sheet to a TSV file.

        Launch a file dialog to determine where and under what name to
        save the TSV file and then write the TSV file to that path.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        dialog = FileDialog(
            parent=self,
            path=str(self.data_path),
            file_mode=QFileDialog.AnyFile,
            accept_mode=QFileDialog.AcceptSave,
            options=QFileDialog.DontConfirmOverwrite,
        )
        if not dialog.exec() == QDialog.Accepted:
            return
        save_path = dialog.selectedFiles()[0]

        # setDefaultSuffix is preferred, but it is not functioning.
        # https://bugreports.qt.io/browse/QTBUG-20011
        if not save_path.endswith(".tsv"):
            save_path += ".tsv"

        if Path(save_path).exists():
            overwrite_dialog = MessageDialog(
                parent=self,
                icon=QMessageBox.Icon.Question,
                title="file conflict",
                text=(
                    f"A file named '{Path(save_path).name}' already exists in "
                    f"'{Path(save_path).parent.name}'.\n"
                    f"Would you like to overwrite '{Path(save_path).name}'?"
                ),
                info="The overwritten file will not be recoverable.",
                details=f"Full path of conflicting file:\n{save_path}",
                buttons=(QMessageBox.No | QMessageBox.Yes),
                default=QMessageBox.No,
            )
            if not overwrite_dialog.exec() == QMessageBox.Yes:
                status = "Export cancelled."
                self.statusbar.showMessage(status, self.statusbar.timeout)
                return
            else:
                Path(save_path).unlink()

        with open(save_path, "w", newline='') as tsv_file:
            tsv_writer = csv.writer(
                tsv_file,
                delimiter="\t",
                quotechar="|",
                quoting=csv.QUOTE_MINIMAL,
            )
            for row_index in range(self.model.rowCount()):
                record_data = []
                record = self.model.record(row_index)
                for column_index in range(record.count()):
                    record_data.append(record.value(column_index))
                tsv_writer.writerow(record_data)

        status = f"Successfully exported current sheet to '{save_path}'."
        self.statusbar.showMessage(status, self.statusbar.timeout)

    def data_import(self) -> None:
        """Import a TSV file into the current sheet.

        Launch a file dialog to determine what TSV file to open. The
        records contained in the TSV file are appended to the current
        sheet.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        dialog = FileDialog(
            parent=self,
            path=str(self.data_path),
            file_mode=QFileDialog.ExistingFile,
            accept_mode=QFileDialog.AcceptOpen,
            name_filter="Tab-separated value (*.tsv)",
        )
        if not dialog.exec() == QDialog.Accepted:
            return
        open_path = dialog.selectedFiles()[0]

        with open(open_path, newline="") as tsv_file:
            tsv_reader = csv.reader(
                tsv_file,
                delimiter="\t",
                quotechar="|",
            )
            row_count = 0
            for row in tsv_reader:
                record = self.model.record(-1)
                for index, item in enumerate(row):
                    if record.fieldName(index) == "id":
                        continue
                    record.setValue(index, item)
                self.model.insertRecord(-1, record)
                row_count += 1

        status = f"Import successful. {row_count} rows inserted."
        self.statusbar.showMessage(status, self.statusbar.timeout)

    def filter_view(self) -> None:
        """Set a filter on the view to only display certain records.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the filter syntax is not correct.

        """
        self.model.setFilter("")
        self.model.select()

        query = self.searchbar.text()
        try:
            sql_string = tablequery.main(query, self.headers)
        except ValueError:
            dialog = MessageDialog(
                parent=self,
                icon=QMessageBox.Icon.Warning,
                title="search error",
                text="Invalid search syntax.".ljust(85),
                info=(
                    "Please check for typos or missing quotes.\n"
                    "For search syntax help, view the 'Help - Help' menu."
                ),
                buttons=QMessageBox.Ok,
            )
            dialog.exec()
            self.model.setFilter("")
            self.model.select()

            status = f"All records displayed ({self.model.rowCount()})."
            self.statusbar.showMessage(status, self.statusbar.timeout)
        else:
            self.model.setFilter(sql_string)
            self.model.select()

            status = f"Search results found ({self.model.rowCount()})."
            self.statusbar.showMessage(status, self.statusbar.timeout)

    def handle_transition(self, old_con_name: str) -> None:
        """Update aspects of application state related to a connection.

        This method handles various tasks necessitated by the transition
        from one connection to another, such as: changing the window
        title, closing the old connection, resetting the model, updating
        the view, and updating the recent files menu.

        Parameters
        ----------
        old_con_name
            The database connection from which to transition.

        Returns
        -------
        None

        """
        self.set_window_title()

        old_con = QSqlDatabase.database(old_con_name)
        old_con.close()
        assert old_con.isOpen() is False

        self._configure_model()
        self.view.setModel(self.model)

        self.update_recent_files()
        self.update_recent_files_menu()

    def insert_rows(
        self,
        checked: bool,
        record: Optional[QSqlRecord] = None,
    ) -> None:
        """Insert a row (record) into the model.

        If a record is not given, an empty record from the existing
        model will be inserted.

        Parameters
        ----------
        checked
            Whether the triggered QAction is checked or not. Although
            the action that calls this method is not checkable, the
            state is nevertheless provided with the signal.
        record
            The record to insert.

        Returns
        -------
        None

        """
        del checked  # Provided with signal, but is not useful.
        # The primeInsert signal that is emitted by insertRecord does
        # not contain row ID information because the row has not been
        # fully initialized in the database yet.
        # To work around this, block signals until the record is fully
        # initialized, then manually emit the signal so the row ID can
        # be obtained (necessary for undo/redo).
        self.model.blockSignals(True)
        if record is None:
            self.model.insertRecord(-1, self.model.record(-1))
        else:
            self.model.insertRecord(-1, record)
        self.model.blockSignals(False)
        self.model.select()

        row_ids = {}
        for index in range(self.model.rowCount()):
            row_ids[self.model.record(index).value("id")] = index
        new_index = row_ids[max(row_ids.keys())]
        new_record = self.model.record(new_index)
        self.model.primeInsert.emit(new_index, new_record)  # type: ignore

        if self.settings.value("table/scroll_to_inserted") == "0":
            pass
        else:
            self.make_row_visible(new_index)
            self.view.selectRow(new_index)

        status = f"Row inserted at row {new_index + 1}."
        self.statusbar.showMessage(status, self.statusbar.timeout)

    def is_state_resolved(self) -> bool:
        """Ensure that unsaved changes are not lost before a transition.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        db_name = QSqlDatabase.database(self.con_name).databaseName()

        if (
            (db_name == ":memory:" and self.model.rowCount() > 0)
            or self.model.isDirty() is True
        ):
            dialog = MessageDialog(
                parent=self,
                icon=QMessageBox.Icon.Question,
                title="save changes",
                text="Would you like to save changes before closing?",
                info="Unsaved changes will be lost after closing.",
                buttons=(
                    QMessageBox.Discard |
                    QMessageBox.Cancel |
                    QMessageBox.Save
                ),
                default=QMessageBox.Cancel,
            )
            choice = dialog.exec()
        else:
            return True

        if db_name == ":memory:" and choice == QMessageBox.Save:
            self.save_as()
            return True
        elif self.model.isDirty() is True and choice == QMessageBox.Save:
            self.save_model()
            return True
        elif choice == QMessageBox.Discard:
            return True
        else:
            return False

    def launch_about_window(self) -> None:
        """Open a window that displays application information.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        # If window is open, do not re-create when this method called.
        if self.about_window is None:
            self.about_window = information.AboutWindow(self)
        else:
            self.about_window.activateWindow()
        self.about_window.show()

    def launch_help_window(self) -> None:
        """Open a window that displays application help information.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        # If window is open, do not re-create when this method called.
        if self.help_window is None:
            self.help_window = information.HelpWindow(self)
        else:
            self.help_window.activateWindow()
        self.help_window.show()

    def launch_metadata_window(self) -> None:
        """Open a window that can be used to retrieve metadata.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        # If window is open, do not re-create when this method called.
        if self.metadata_window is None:
            self.metadata_window = metadata.MetadataWindow(self)
        else:
            self.metadata_window.activateWindow()
        self.metadata_window.show()

    def load_settings(self) -> None:
        """Restore the application state stored during the last run.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        window_geometry = self.settings.value("window/geometry")
        if window_geometry is not None:
            self.restoreGeometry(window_geometry)

        header_state = self.settings.value("header/state")
        if header_state is not None:
            self.header.restoreState(header_state)
            self.view.updateGeometries()

        statusbar_state = self.settings.value("statusbar/visible")
        if statusbar_state is None or statusbar_state == "1":
            self.statusbar.setVisible(True)
        else:
            self.statusbar.setVisible(False)

    def make_row_visible(self, row_index: int) -> None:
        """Ensure a particular row is made visible in the viewport.

        Parameters
        ----------
        row_index
            The row to be made visible.

        Returns
        -------
        None

        """
        for visual_index in range(self.model.columnCount()):
            logical_index = self.header.logicalIndex(visual_index)
            if not self.header.isSectionHidden(logical_index):
                first_visible = logical_index
                break
        self.view.scrollTo(self.model.createIndex(row_index, first_visible))

    def new_sheet(self) -> None:
        """Open a new, in-memory sheet.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        if not self.is_state_resolved():
            return

        old_con_name = self.con_name
        self.set_new_connection_name("virtual")
        open_db(self.con_name, ":memory:")
        create_empty_table(self.con_name)

        self.handle_transition(old_con_name)

        status = "Opened new sheet."
        self.statusbar.showMessage(status, self.statusbar.timeout)

    def open_recent_file(self, db_path: str) -> None:
        """Open a disk-based database.

        Parameters
        ----------
        db_path
            The path to the file to open.

        Returns
        -------
        None

        """
        if not self.is_state_resolved():
            return

        if not Path(db_path).exists():
            MessageDialog(
                parent=self,
                icon=QMessageBox.Icon.Warning,
                title="open recent file error",
                text=(
                    "The file could not be opened because it has\n"
                    "either been moved or deleted."
                ),
                details=f"Full path of the original file:\n{db_path}",
                buttons=QMessageBox.Ok,
            ).exec()
            return

        old_con_name = self.con_name
        self.set_new_connection_name("physical")
        open_db(self.con_name, db_path)

        self.handle_transition(old_con_name)

        status = f"Opened '{db_path}'."
        self.statusbar.showMessage(status, self.statusbar.timeout)

    def open_sheet(self) -> None:
        """Open an existing, disk-based database.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        if not self.is_state_resolved():
            return
        dialog = FileDialog(
            parent=self,
            path=str(self.data_path),
            file_mode=QFileDialog.ExistingFile,
            accept_mode=QFileDialog.AcceptOpen,
            name_filter="SQLite3 database (*.sqlite)",
        )
        if not dialog.exec() == QDialog.Accepted:
            return
        open_path = dialog.selectedFiles()[0]

        old_con_name = self.con_name
        self.set_new_connection_name("physical")
        open_db(self.con_name, open_path)

        self.handle_transition(old_con_name)

        status = f"Opened '{open_path}'."
        self.statusbar.showMessage(status, self.statusbar.timeout)

    def perform_redo(self) -> None:
        """Redo an undone row operation (insertion or deletion).

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        redo_info = self.model.redo_stack.pop()
        if redo_info["command"] == "deletion":
            # Find the record that needs to be deleted based on ID.
            for row_index in range(self.model.rowCount()):
                record_id = self.model.record(row_index).value("id")
                if record_id == redo_info["record_id"]:
                    old_record = cast(QSqlRecord, redo_info["record"])
                    # Save changes to the record before deleting it.
                    updated_record = self.update_record(
                        old_record,
                        self.model.record(row_index),
                    )
                    redo_info["record"] = updated_record
                    self.model.undo_stack.append(redo_info)

                    # Ensure this redo does not trigger an undo.
                    self.model.blockSignals(True)
                    self.model.removeRows(row_index, 1)
                    self.model.blockSignals(False)

                    status = "Redid row deletion."
                    self.statusbar.showMessage(status, self.statusbar.timeout)
        if redo_info["command"] == "insertion":
            deleted_record = cast(QSqlRecord, redo_info["record"])

            # Ensure this redo does not trigger an undo.
            self.model.blockSignals(True)
            self.model.insertRecord(-1, deleted_record)
            self.model.blockSignals(False)

            # Update old references to this record to include new ID.
            reinserted_record = self.model.record(self.model.rowCount() - 1)
            reinserted_record_id = reinserted_record.value("id")
            self.refresh_redo_stacks(
                redo_info["record_id"],
                reinserted_record_id,
            )
            redo_info["record_id"] = reinserted_record_id
            self.model.undo_stack.append(redo_info)

            status = "Redid row insertion."
            self.statusbar.showMessage(status, self.statusbar.timeout)
        self.model.select()
        self.toggle_undo_enable()
        self.toggle_redo_enable()

    def perform_undo(self) -> None:
        """Undo a row operation (insertion or deletion).

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        undo_info = self.model.undo_stack.pop()
        if undo_info["command"] == "deletion":
            deleted_record = cast(QSqlRecord, undo_info["record"])

            # Ensure this undo does not trigger another undo.
            self.model.blockSignals(True)
            self.model.insertRecord(-1, deleted_record)
            self.model.blockSignals(False)

            # Update old references to this record to include new ID.
            undeleted_record = self.model.record(self.model.rowCount() - 1)
            undeleted_record_id = undeleted_record.value("id")
            self.refresh_undo_stacks(
                undo_info["record_id"],
                undeleted_record_id,
            )
            undo_info["record_id"] = undeleted_record_id
            self.model.redo_stack.append(undo_info)

            status = "Undid row deletion."
            self.statusbar.showMessage(status, self.statusbar.timeout)
        if undo_info["command"] == "insertion":
            # Find the record that needs to be deleted based on ID.
            for row_index in range(self.model.rowCount()):
                record_id = self.model.record(row_index).value("id")
                if record_id == undo_info["record_id"]:
                    old_record = cast(QSqlRecord, undo_info["record"])
                    # Save changes to the record before deleting it.
                    updated_record = self.update_record(
                        old_record,
                        self.model.record(row_index),
                    )
                    undo_info["record"] = updated_record
                    self.model.redo_stack.append(undo_info)

                    # Ensure this undo does not trigger another undo.
                    self.model.blockSignals(True)
                    self.model.removeRows(row_index, 1)
                    self.model.blockSignals(False)

                    status = "Undid row insertion."
                    self.statusbar.showMessage(status, self.statusbar.timeout)
        self.model.select()
        self.toggle_undo_enable()
        self.toggle_redo_enable()

    def refresh_redo_stacks(
        self,
        old_record_id: Any,
        new_record_id: int,
    ) -> None:
        """Update references to the old record in the redo stack.

        Parameters
        ----------
        old_record_id
            The ID of the record before being re-inserted.
        new_record_id
            The ID of the record after being re-inserted.

        Returns
        -------
        None

        Notes
        -----
        Records are deleted based on the ID that is held in the stack.
        When a record is re-inserted it receives a new ID. Therefore,
        if it is desired to undo the record's insertion, all references
        to that record that exist in the stack must be made to reflect
        the record's new ID.

        """
        for redo in self.model.redo_stack:
            if redo["record_id"] == old_record_id:
                redo["record_id"] = new_record_id

    def refresh_undo_stacks(
        self,
        old_record_id: Any,
        new_record_id: int,
    ) -> None:
        """Update references to the old record in the undo stack.

        Parameters
        ----------
        old_record_id
            The ID of the record before being re-inserted.
        new_record_id
            The ID of the record after being re-inserted.

        Returns
        -------
        None

        Notes
        -----
        Records are deleted based on the ID that is held in the stack.
        When a record is re-inserted it receives a new ID. Therefore,
        if it is desired to undo the record's insertion, all references
        to that record that exist in the stack must be made to reflect
        the record's new ID.

        """
        for undo in self.model.undo_stack:
            if undo["record_id"] == old_record_id:
                undo["record_id"] = new_record_id

    def remove_rows(self) -> None:
        """Remove the currently-selected rows.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        removed_indexes = []
        for qt_index in self.view.selectedIndexes():
            # If block selection, do not delete a row multiple times.
            if qt_index.row() not in removed_indexes:
                self.model.removeRows(qt_index.row(), 1)
                removed_indexes.append(qt_index.row())
        self.model.select()

        removed_rows = [index + 1 for index in removed_indexes]
        if len(removed_rows) == 1:
            status = f"Row {removed_rows[0]} removed."
        elif len(removed_rows) > 1:
            removed_rows.sort()
            min_row = removed_rows[0]
            max_row = removed_rows[-1]
            if removed_rows == [i for i in range(min_row, max_row + 1)]:
                status = f"Rows {min_row}-{max_row} removed."
            else:
                formatted = ", ".join([str(i) for i in removed_rows])
                status = f"Rows {formatted} removed."
        else:
            status = "No rows removed. At least one row must be selected."
        self.statusbar.showMessage(status, self.statusbar.timeout)

    def reset_filter(self, current_text: str) -> None:
        """Clear the filter and show all records when no searchbar text.

        Parameters
        ----------
        current_text
            The text retrieved from the searchbar.

        Returns
        -------
        None

        """
        if current_text == "":
            self.model.setFilter("")
            self.model.select()

            status = f"All records displayed ({self.model.rowCount()})."
            self.statusbar.showMessage(status, self.statusbar.timeout)

    def resize_and_center(self) -> None:
        """Resize and center the window to create a default state.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        # Can technically return None, but appease mypy that it won't.
        app = QApplication.instance()
        assert app is not None
        screen = app.primaryScreen()

        screen_width = screen.availableGeometry().width()
        screen_height = screen.availableGeometry().height()
        window_width = int(screen_width * 0.8)
        window_height = int(screen_height * 0.8)
        self.resize(window_width, window_height)

        screen_center = screen.geometry().center()
        frame_center = self.frameGeometry().center()
        self.move(screen_center - frame_center)

    def restore_default_header_state(self) -> None:
        """Reconfigure the header to its default state.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        header_default_state = self.settings.value("header/default_state")
        if header_default_state is not None:
            self.header.restoreState(header_default_state)
            self.view.updateGeometries()
            self.set_header_check_states()

            status = "Default column configuration restored."
            self.statusbar.showMessage(status, self.statusbar.timeout)

    def restore_default_window_geometry(self) -> None:
        """Reconfigure the window geometry to its default state.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        window_geometry = self.settings.value("window/default_geometry")
        if window_geometry is not None:
            self.restoreGeometry(window_geometry)

            status = "Default window configuration restored."
            self.statusbar.showMessage(status, self.statusbar.timeout)

    def save_as(self) -> None:
        """Save the current sheet to the filesystem.

        Given user confirmation, the method has the ability to overwrite
        files, though it cannot overwrite the file that is currently
        open. Overwritten files are not retrievable.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Raises
        ------
        send2trash.TrashPermissionError
            If the device does not have .Trash and no permission to
            create it.

        """
        dialog = FileDialog(
            parent=self,
            path=str(self.data_path),
            file_mode=QFileDialog.AnyFile,
            accept_mode=QFileDialog.AcceptSave,
            options=QFileDialog.DontConfirmOverwrite,
        )
        if not dialog.exec() == QDialog.Accepted:
            return
        save_path = dialog.selectedFiles()[0]

        # setDefaultSuffix is preferred, but it is not functioning.
        # https://bugreports.qt.io/browse/QTBUG-20011
        if not save_path.endswith(".sqlite"):
            save_path += ".sqlite"

        if Path(save_path).exists():
            overwrite_dialog = MessageDialog(
                parent=self,
                icon=QMessageBox.Icon.Question,
                title="file conflict",
                text=(
                    f"A file named '{Path(save_path).name}' already exists in "
                    f"'{Path(save_path).parent.name}'.\n"
                    f"Would you like to overwrite '{Path(save_path).name}'?"
                ),
                info="The overwritten file will moved to the trash.",
                details=f"Full path of conflicting file:\n{save_path}",
                buttons=(QMessageBox.No | QMessageBox.Yes),
                default=QMessageBox.No,
            )
            if not overwrite_dialog.exec() == QMessageBox.Yes:
                status = "Save cancelled."
                self.statusbar.showMessage(status, self.statusbar.timeout)
                return
            else:
                # User wants to delete, but one more check must be done.
                deletion_flag = True
        else:
            deletion_flag = False

        current_db_name = QSqlDatabase.database(self.con_name).databaseName()
        if deletion_flag and save_path == current_db_name:
            MessageDialog(
                parent=self,
                icon=QMessageBox.Icon.Warning,
                title="save error",
                text="The file that is currently open cannot be overwritten.",
                info="Please save to a different file name.",
                buttons=QMessageBox.Ok,
            ).exec()
            status = "Save cancelled."
            self.statusbar.showMessage(status, self.statusbar.timeout)
            return

        if deletion_flag:
            try:
                send2trash.send2trash(save_path)
            except send2trash.TrashPermissionError:
                MessageDialog(
                    parent=self,
                    icon=QMessageBox.Icon.Warning,
                    title="save error",
                    text=(
                        "The existing file was unable to be sent to the trash."
                    ),
                    info="Please move the existing file and try again.",
                    buttons=QMessageBox.Ok,
                ).exec()
                status = "Save cancelled."
                self.statusbar.showMessage(status, self.statusbar.timeout)
                return

        old_con_name = self.con_name
        self.set_new_connection_name("physical")
        open_db(self.con_name, save_path)

        create_empty_table(self.con_name)
        copy_all_data(old_con_name, self.con_name)

        self.handle_transition(old_con_name)

        status = f"Saved file as '{save_path}'."
        self.statusbar.showMessage(status, self.statusbar.timeout)

    def save_default_header_state(self) -> None:
        """Save header config before loading custom config.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        header_state = self.header.saveState()
        self.settings.setValue("header/default_state", header_state)

    def save_default_window_geometry(self) -> None:
        """Save window config before loading custom config.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        window_geometry = self.saveGeometry()
        self.settings.setValue("window/default_geometry", window_geometry)

    def save_header_state(self) -> None:
        """Save the current header state.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        header_state = self.header.saveState()
        self.settings.setValue("header/state", header_state)

    def save_model(self) -> None:
        """Save the model.

        Models that exist in-memory will either be transitioned to a
        disk-based database or discarded.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        if QSqlDatabase.database(self.con_name).databaseName() == ":memory:":
            self.save_as()
        else:
            self.model.submitAll()
            self.model.select()

            db_path = QSqlDatabase.database(self.con_name).databaseName()
            status = f"Saved current file '{db_path}'."
            self.statusbar.showMessage(status, self.statusbar.timeout)

    def save_window_geometry(self) -> None:
        """Save the current window geometry.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        window_geometry = self.saveGeometry()
        self.settings.setValue("window/geometry", window_geometry)

    def set_header_check_states(self) -> None:
        """Update the check states of header actions.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Notes
        -----
        Header check states describe whether or not a particular header
        is hidden or not.

        """
        for index, context_action in enumerate(self.header_context_actions):
            if not self.header.isSectionHidden(index):
                context_action.setChecked(True)
            else:
                context_action.setChecked(False)

    def set_new_connection_name(self, con_type: str) -> None:
        """Set a new database connection name.

        Parameters
        ----------
        con_type
            The type of connection (virtual or physical) to set.

        Returns
        -------
        None

        Notes
        -----
        To prevent unintended connection reuse, the highest numbered
        existing connection is found and incremented.

        """
        con_names = QSqlDatabase.database().connectionNames()
        sep_con_names = list(
            filter(
                lambda s: s.startswith("{}_".format(con_type)),
                con_names,
            ),
        )
        if sep_con_names:
            con_numbers = [int(s[s.index("_") + 1:]) for s in sep_con_names]
            self.con_name = "{}_{}".format(con_type, max(con_numbers) + 1)
        else:
            self.con_name = "{}_1".format(con_type)

    def set_optimal_widths(self) -> None:
        """Resize columns to fit their contents.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        self.header.setSectionResizeMode(QHeaderView.ResizeToContents)
        self.view.updateGeometries()
        self.header.setSectionResizeMode(QHeaderView.Interactive)

        status = "Columns resized to optimal widths."
        self.statusbar.showMessage(status, self.statusbar.timeout)

    def set_preferences(self) -> None:
        """Provide access to various settings.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        dialog = PreferencesDialog(self)
        if dialog.exec():
            scroll = dialog.scroll_checkbox.isChecked()
            if scroll:
                self.settings.setValue("table/scroll_to_inserted", "1")
            else:
                self.settings.setValue("table/scroll_to_inserted", "0")

            # The statusbar uses milliseconds instead of seconds.
            duration = dialog.duration_spinbox.value() * 1000
            self.statusbar.timeout = duration
            self.settings.setValue("statusbar/duration", duration)

            visible = dialog.visible_checkbox.isChecked()
            self.statusbar.setVisible(visible)
            if visible is True:
                self.settings.setValue("statusbar/visible", "1")
            else:
                self.settings.setValue("statusbar/visible", "0")
            status = "Preferences saved."
        else:
            status = "Preferences not saved."
        self.statusbar.showMessage(status, self.statusbar.timeout)

    def set_window_title(self) -> None:
        """Update the window title to include database information.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        con = QSqlDatabase.database(self.con_name)
        db_name = con.databaseName()
        if db_name == ":memory:":
            self.setWindowTitle("New - filmbuff")
        else:
            self.setWindowTitle("{} - filmbuff".format(Path(db_name).stem))

    def toggle_header_visibility(self, logical_index: int) -> None:
        """Toggle the visbility of a header.

        Parameters
        ----------
        logical_index
            The absolute position of the header in the QHeaderView.

        Returns
        -------
        None

        """
        header_name = self.model.headerData(logical_index, Qt.Horizontal)
        if self.header.isSectionHidden(logical_index):
            self.header.showSection(logical_index)
            status = f"Column '{header_name}' is now shown."
            self.statusbar.showMessage(status, self.statusbar.timeout)
        else:
            self.header.hideSection(logical_index)
            status = f"Column '{header_name}' is now hidden."
            self.statusbar.showMessage(status, self.statusbar.timeout)

    def toggle_redo_enable(self) -> None:
        """Enable or disable the redo action.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        if self.model.redo_stack:
            self.redo_action.setEnabled(True)
        else:
            self.redo_action.setEnabled(False)

    def toggle_undo_enable(self) -> None:
        """Enable or disable the undo action.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        if self.model.undo_stack:
            self.undo_action.setEnabled(True)
        else:
            self.undo_action.setEnabled(False)

    def update_recent_files(self) -> None:
        """Update the list of recent files stored in the settings.

        Parameters
        ----------
        None

        Returns
        -------
        None

        See Also
        --------
        update_recent_files_menu

        """
        db_path = Path(QSqlDatabase.database(self.con_name).databaseName())

        recent_files = self.settings.value("recent_files")
        if recent_files is None:
            recent_files = []

        if str(db_path) == ":memory:":
            return
        elif db_path in recent_files:
            # Float an old entry to the top of the list.
            del recent_files[recent_files.index(db_path)]
            recent_files.append(db_path)
        else:
            recent_files.append(db_path)

        if len(recent_files) > 6:
            # Trim the oldest entry from the list.
            recent_files.pop(0)

        self.settings.setValue("recent_files", recent_files)

    def update_recent_files_menu(self) -> None:
        """Update the recent files menu based on the settings.

        Parameters
        ----------
        None

        Returns
        -------
        None

        See Also
        --------
        update_recent_files

        """
        recent_files = self.settings.value("recent_files")
        if recent_files is None:
            return

        self.recent_files_actions = []
        # The first file in the list is the oldest file.
        # The last file in the list is the current file (not displayed).
        for index in range(len(recent_files) - 2, -1, -1):
            self.recent_file_action = QAction()
            self.recent_file_action.setText(str(recent_files[index].name))
            self.recent_file_action.triggered.connect(
                lambda _, i=index: self.open_recent_file(str(recent_files[i])),
            )
            self.recent_files_menu.addAction(self.recent_file_action)
            self.recent_files_actions.append(self.recent_file_action)

    def update_undo_stack(self, description: str, index: int) -> None:
        """Add an item to the undo stack.

        Parameters
        ----------
        description
            The type of command performed (deletion or insertion).
        index
            The row on which the operation was performed.

        Returns
        -------
        None

        """
        if description == "deletion":
            undo_info = {
                "command": "deletion",
                "record": self.model.record(index),
                "record_id": self.model.record(index).value("id"),
            }
            self.model.undo_stack.append(undo_info)
        if description == "insertion":
            undo_info = {
                "command": "insertion",
                "record": self.model.record(index),
                "record_id": self.model.record(index).value("id"),
            }
            self.model.undo_stack.append(undo_info)

        self.toggle_undo_enable()
        self.toggle_redo_enable()

    @staticmethod
    def make_data_path() -> Path:
        """Ensure a standard location to store data exists.

        Parameters
        ----------
        None

        Returns
        -------
        :obj:`pathlib.Path`
            The standard location for storing persistent data.

        """
        path_enum = QStandardPaths.GenericDataLocation
        parent_path = Path(QStandardPaths.writableLocation(path_enum))
        data_path = parent_path / QCoreApplication.applicationName()
        data_path.mkdir(exist_ok=True)
        return data_path

    @staticmethod
    def update_record(outdated: QSqlRecord, updated: QSqlRecord) -> QSqlRecord:
        """Update a record's fields to match another record's field.

        Parameters
        ----------
        outdated
            The record to update.
        updated
            The record containing the desired values.

        Returns
        -------
        :obj:`PyQt5.QtQSql.QSqlRecord`
            The outdated record with all fields (except 'ID') updated.

        Notes
        -----
        The 'ID' field is not updated because it is automatically
        generated by the database.

        """
        # Overwriting the old record with the new record makes it not
        # able to be inserted. Instead, make the old record contain the
        # same information as the new record.
        for column_index in range(updated.count()):
            if updated.fieldName(column_index) == "ID":
                continue
            outdated.setValue(column_index, updated.value(column_index))
        return outdated

    def _configure_actions(self) -> None:
        self.restore_window_action = QAction()
        self.restore_window_action.setText("Window configuration")
        self.restore_window_action.triggered.connect(
            self.restore_default_window_geometry,
        )

        self.restore_header_action = QAction()
        self.restore_header_action.setText("Column configuration")
        self.restore_header_action.triggered.connect(
            self.restore_default_header_state,
        )

        self.header_context_actions = []
        for index, header in enumerate(self.headers.keys()):
            self.header_context_action = QAction()
            self.header_context_action.setCheckable(True)
            self.header_context_action.setText(header)
            self.header_context_action.triggered.connect(
                lambda check_state, i=index: self.toggle_header_visibility(i),
            )
            self.header_context_actions.append(self.header_context_action)

        self.optimal_widths_action = QAction()
        self.optimal_widths_action.setText("Optimal widths")
        self.optimal_widths_action.triggered.connect(self.set_optimal_widths)

        self.new_action = QAction()
        self.new_action.setText("New")
        self.new_action.setIcon(QIcon.fromTheme("document-new"))
        self.new_action.setShortcut(QKeySequence("Ctrl+n"))
        self.new_action.triggered.connect(self.new_sheet)

        self.open_action = QAction()
        self.open_action.setText("Open...")
        self.open_action.setIcon(QIcon.fromTheme("document-open"))
        self.open_action.setShortcut(QKeySequence("Ctrl+o"))
        self.open_action.triggered.connect(self.open_sheet)

        self.save_action = QAction()
        self.save_action.setText("Save")
        self.save_action.setIcon(QIcon.fromTheme("document-save"))
        self.save_action.setShortcut(QKeySequence("Ctrl+s"))
        self.save_action.triggered.connect(self.save_model)

        self.save_as_action = QAction()
        self.save_as_action.setText("Save as...")
        self.save_as_action.setIcon(QIcon.fromTheme("document-save-as"))
        self.save_as_action.setShortcut(QKeySequence("Shift+Ctrl+s"))
        self.save_as_action.triggered.connect(self.save_as)

        self.import_action = QAction()
        self.import_action.setText("Import tsv...")
        self.import_action.triggered.connect(self.data_import)

        self.export_action = QAction()
        self.export_action.setText("Export tsv...")
        self.export_action.triggered.connect(self.data_export)

        self.exit_action = QAction()
        self.exit_action.setText("Exit")
        self.exit_action.setShortcut(QKeySequence("Ctrl+q"))
        self.exit_action.triggered.connect(self.close)

        self.undo_action: QAction = QAction()
        self.undo_action.setText("Undo")
        self.undo_action.setIcon(QIcon.fromTheme("edit-undo"))
        self.undo_action.setShortcut(QKeySequence("Ctrl+z"))
        self.undo_action.triggered.connect(self.perform_undo)
        self.toggle_undo_enable()

        self.redo_action: QAction = QAction()
        self.redo_action.setText("Redo")
        self.redo_action.setIcon(QIcon.fromTheme("edit-redo"))
        self.redo_action.setShortcut(QKeySequence("Ctrl+y"))
        self.redo_action.triggered.connect(self.perform_redo)
        self.toggle_redo_enable()

        self.add_action = QAction()
        self.add_action.setText("Add row")
        self.add_action.setIcon(QIcon.fromTheme("list-add"))
        self.add_action.setShortcut(QKeySequence("Ctrl++"))
        self.add_action.triggered.connect(self.insert_rows)

        self.remove_action = QAction()
        self.remove_action.setText("Remove selected row(s)")
        self.remove_action.setIcon(QIcon.fromTheme("list-remove"))
        self.remove_action.setShortcut(QKeySequence("Ctrl+-"))
        self.remove_action.triggered.connect(self.remove_rows)

        self.preferences_action = QAction()
        self.preferences_action.setText("Preferences...")
        self.preferences_action.setShortcut(QKeySequence("Ctrl+p"))
        self.preferences_action.triggered.connect(self.set_preferences)

        self.help_action = QAction()
        self.help_action.setText("Help")
        self.help_action.setShortcut(QKeySequence("F1"))
        self.help_action.triggered.connect(self.launch_help_window)

        self.about_action = QAction()
        self.about_action.setText("About")
        self.about_action.triggered.connect(self.launch_about_window)

        self.search_action = QAction()
        self.search_action.setText("Search")
        self.search_action.setIcon(QIcon.fromTheme("system-search"))
        self.search_action.triggered.connect(self.filter_view)

        self.metadata_action = QAction()
        self.metadata_action.setText("Retrieve metadata")
        self.metadata_action.setIcon(QIcon.fromTheme("applications-internet"))
        self.metadata_action.setShortcut(QKeySequence("Ctrl+m"))
        self.metadata_action.triggered.connect(self.launch_metadata_window)

    def _configure_header(self) -> None:
        self.header: QHeaderView = self.view.horizontalHeader()
        self.header.setContextMenuPolicy(Qt.CustomContextMenu)
        self.header.setSortIndicatorShown(True)
        self.header.setSectionsMovable(True)

        self.header.setSectionResizeMode(QHeaderView.ResizeToContents)
        self.header.setStretchLastSection(True)

        self.header.setSectionResizeMode(QHeaderView.Interactive)

    def _configure_header_context(self) -> None:
        self.header_context = QMenu(self)
        self.header_context.addAction(self.optimal_widths_action)
        self.header_context.addSeparator()
        self.header.customContextMenuRequested.connect(
            lambda: self.header_context.exec(QCursor.pos()),
        )
        for context_action in self.header_context_actions:
            self.header_context.addAction(context_action)

    def _configure_menubar(self) -> None:
        self.menubar = self.menuBar()
        self.menubar.setContextMenuPolicy(Qt.PreventContextMenu)
        self.setMenuBar(self.menubar)

        self.file_menu = self.menubar.addMenu("&File")
        self.file_menu.addAction(self.new_action)
        self.file_menu.addAction(self.open_action)
        self.recent_files_menu: QMenu = self.file_menu.addMenu("Recent files")
        self.file_menu.addSeparator()
        self.file_menu.addAction(self.save_action)
        self.file_menu.addAction(self.save_as_action)
        self.file_menu.addSeparator()
        self.file_menu.addAction(self.import_action)
        self.file_menu.addAction(self.export_action)
        self.file_menu.addSeparator()
        self.file_menu.addAction(self.exit_action)

        self.edit_menu = self.menubar.addMenu("&Edit")
        self.edit_menu.addAction(self.undo_action)
        self.edit_menu.addAction(self.redo_action)

        self.sheet_menu = self.menubar.addMenu("&Sheet")
        self.sheet_menu.addAction(self.add_action)
        self.sheet_menu.addAction(self.remove_action)
        self.sheet_menu.addSeparator()
        self.sheet_menu.addAction(self.metadata_action)

        self.view_menu = self.menubar.addMenu("&View")
        self.view_resize_menu = self.view_menu.addMenu("Resize columns")
        self.view_resize_menu.addAction(self.optimal_widths_action)
        self.view_menu.addSeparator()
        self.view_columns_menu = self.view_menu.addMenu("Show/hide columns")
        for action in self.header_context_actions:
            self.view_columns_menu.addAction(action)
        self.view_menu.addSeparator()
        self.view_defaults_menu = self.view_menu.addMenu("Restore defaults")
        self.view_defaults_menu.addAction(self.restore_header_action)
        self.view_defaults_menu.addAction(self.restore_window_action)

        self.tools_menu = self.menubar.addMenu("&Tools")
        self.tools_menu.addAction(self.preferences_action)

        self.help_menu = self.menubar.addMenu("&Help")
        self.help_menu.addAction(self.help_action)
        self.help_menu.addSeparator()
        self.help_menu.addAction(self.about_action)

    def _configure_model(self) -> None:
        self.model = Model(QSqlDatabase.database(self.con_name), self)
        # Reformat model headers to be more readable.
        self.headers = {}
        for index in range(self.model.columnCount()):
            model_header = self.model.headerData(index, Qt.Horizontal)

            display_header = model_header.replace("_", " ")
            if "id" in display_header.lower():
                display_header = display_header.upper()
            else:
                display_header = display_header.capitalize()

            self.headers[display_header] = model_header
            self.model.setHeaderData(index, Qt.Horizontal, display_header)
        self.model.select()

    def _configure_toolbar(self) -> None:
        self.spacer = QWidget()
        self.spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self.searchbar = QLineEdit()
        self.searchbar.setClearButtonEnabled(True)
        self.searchbar.setPlaceholderText(
            "Type here and press enter to search...",
        )
        self.searchbar.returnPressed.connect(self.filter_view)
        self.searchbar.textChanged.connect(self.reset_filter)

        self.toolbar = QToolBar()
        self.toolbar.setFloatable(False)
        self.addToolBar(self.toolbar)
        self.toolbar.setMovable(False)
        self.toolbar.setContextMenuPolicy(Qt.PreventContextMenu)

        self.toolbar.addAction(self.new_action)
        self.toolbar.addAction(self.open_action)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.save_action)
        self.toolbar.addAction(self.save_as_action)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.undo_action)
        self.toolbar.addAction(self.redo_action)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.add_action)
        self.toolbar.addAction(self.remove_action)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.metadata_action)
        self.toolbar.addSeparator()
        self.toolbar.addWidget(self.spacer)
        self.toolbar.addWidget(self.searchbar)
        self.toolbar.addAction(self.search_action)

    def _configure_view(self) -> None:
        self.view = QTableView()
        self.view.setModel(self.model)
        self.view.setAlternatingRowColors(True)
        self.view.setSortingEnabled(True)


class MessageDialog(QMessageBox):
    """A configurable QMessageBox for user interaction.

    Intended use cases include: notifying the user of an error, asking
    the user for a decision, etc.

    """

    def __init__(self, **kwargs: Any) -> None:
        """Construct a MessageDialog instance.

        Parameters
        ----------
        **kwargs
            A dictionary of QMessageBox options.
            Each key corresponds to a QMessageBox method with that key's
            value being directly forwarded to the corresponding method.
            The key's and their corresponding methods are listed below:
                parent - __init__
                icon - setIcon
                title - setWindowTitle
                text - setText
                info - setInformativeText
                details - setDetailedText
                buttons - setStandardButtons
                default - setDefaultButton


        Returns
        -------
        None

        """
        super().__init__(kwargs.get("parent"))
        if kwargs.get("icon") is not None:
            self.setIcon(kwargs["icon"])
        if kwargs.get("title") is not None:
            self.setWindowTitle(f"filmbuff - {kwargs['title']}")
        if kwargs.get("text") is not None:
            self.setText(kwargs["text"])
        if kwargs.get("info") is not None:
            self.setInformativeText(kwargs["info"])
        if kwargs.get("details") is not None:
            self.setDetailedText(kwargs["details"])
        if kwargs.get("buttons") is not None:
            self.setStandardButtons(kwargs["buttons"])
        if kwargs.get("default") is not None:
            self.setDefaultButton(kwargs["default"])


class Model(QSqlTableModel):
    """An editable data model for a single database table."""

    def __init__(self, db: QSqlDatabase, parent: QMainWindow) -> None:
        """Construct a Model instance.

        Parameters
        ----------
        db
            The currently open database connection.
        parent
            The model's parent.

        Returns
        -------
        None

        """
        super().__init__(parent, db)
        self.setTable("films")
        self.setEditStrategy(self.OnRowChange)
        self.undo_stack: List[Dict[str, Union[str, QSqlRecord, int]]] = []
        self.redo_stack: List[Dict[str, Union[str, QSqlRecord, int]]] = []

        self.beforeDelete.connect(  # type: ignore
            lambda row: parent.update_undo_stack("deletion", row),
        )
        self.primeInsert.connect(  # type: ignore
            lambda row, record: parent.update_undo_stack("insertion", row),
        )

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        """Returns the item flags for the given index.

        Parameters
        ----------
        index
            The index to inspect.

        Returns
        -------
        :obj:`Qt.ItemFlags`
            An enum that describes the properties of an item.

        Notes
        -----
        This is an overridden baseclass (QSqlTableModel) method.

        This method is automatically called anytime an index is
        accessed (by PyQt or otherwise). The return value is not
        directly used, but it does affect the properties of the model's
        items.

        """
        flags: int = 0
        # The ID column (always 10) should not be editable.
        if index.column() == 10:
            flags = Qt.ItemIsSelectable
        else:
            flags = Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsEditable
        return cast(Qt.ItemFlags, flags)


class PreferencesDialog(QDialog):
    """A dialog used to obtain settings preferences."""

    def __init__(self, parent: QMainWindow) -> None:
        """Construct a PreferencesDialog instance.

        Parameters
        ----------
        parent
            The PreferenceDialog's parent. The parent must have a
            settings attribute that can be read.

        Returns
        -------
        None

        """
        super().__init__(parent)
        self.setWindowTitle("Preferences")

        table_label = QLabel("Table")
        table_label.setStyleSheet(
            "font-weight: bold; text-decoration: underline;",
        )
        self.scroll_checkbox = QCheckBox()
        scroll = parent.settings.value("table/scroll_to_inserted")
        if scroll == "0":
            self.scroll_checkbox.setChecked(False)
        else:
            self.scroll_checkbox.setChecked(True)

        statusbar_label = QLabel("Statusbar")
        statusbar_label.setStyleSheet(
            "font-weight: bold; text-decoration: underline;",
        )

        self.visible_checkbox = QCheckBox()
        visible = parent.settings.value("statusbar/visible")
        if visible == "0":
            self.visible_checkbox.setChecked(False)
        else:
            self.visible_checkbox.setChecked(True)

        self.duration_spinbox = QSpinBox()
        duration = parent.settings.value("statusbar/duration")
        if duration is not None:
            formatted_duration = int(int(duration) / 1000)
            self.duration_spinbox.setValue(formatted_duration)
        else:
            formatted_duration = int(parent.statusbar.timeout / 1000)
            self.duration_spinbox.setValue(formatted_duration)
        self.duration_spinbox.setSuffix(" s")

        form_layout = QFormLayout()
        form_layout.addRow(table_label)
        form_layout.addRow("Scroll to inserted rows:", self.scroll_checkbox)
        form_layout.addRow(statusbar_label)
        form_layout.addRow("Statusbar is visible:", self.visible_checkbox)
        form_layout.addRow("Message duration:", self.duration_spinbox)

        apply_button = QPushButton("Apply")
        apply_button.pressed.connect(self.accept)
        cancel_button = QPushButton("Cancel")
        cancel_button.pressed.connect(self.reject)
        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(apply_button)
        buttons_layout.addWidget(cancel_button)

        main_layout = QVBoxLayout()
        main_layout.addLayout(form_layout)
        main_layout.addLayout(buttons_layout)
        self.setLayout(main_layout)


class StatusBar(QStatusBar):
    """A statusbar used to present information."""

    def __init__(self, timeout: int, parent: QMainWindow) -> None:
        """Construct a StatusBar instance.

        Parameters
        ----------
        timeout
            The amount of time (milliseconds) before the message
            disappears.
        parent
            The StatusBar's parent.

        Returns
        -------
        None

        """
        super().__init__(parent)
        self.setSizeGripEnabled(False)
        duration = parent.settings.value("statusbar/duration")
        if duration is not None:
            self.timeout = int(duration)
        else:
            self.timeout = timeout

    def showMessage(self, message: str, timeout: int = 0) -> None:
        """Display a message for a certain length of time.

        Parameters
        ----------
        message
            The message to display.
        timeout
            The amount of time (milliseconds) before the message
            disappears.

        Returns
        -------
        None

        Notes
        -----
        This is an overridden baseclass (QStatusBar) method.

        Overriding this method to automatically include `timeout` when
        it is called means that calls to this method do not need to be
        dynamically created. Instead, the timeout attribute can be
        updated when the setting is changed.

        """
        super().showMessage(message, self.timeout)


def copy_all_data(con1_name: str, con2_name: str) -> None:
    """Copy all data from one QSQLite database to another.

    Parameters
    ----------
    con1_name
        The name of the source QSQLite database connection.
    con2_name
        The name of the destination QSQLite database connection.

    Returns
    -------
    None

    """
    con2 = QSqlDatabase.database(con2_name)
    db2_insertion = QSqlQuery(con2)
    db2_insertion.prepare(
        """
        INSERT INTO films (
            budget,
            cast_one,
            cast_two,
            cast_three,
            color,
            content,
            director,
            genre_one,
            genre_two,
            genre_three,
            imdb_id,
            notes,
            original_language,
            original_title,
            overview,
            popularity,
            production_country,
            release_date,
            revenue,
            runtime,
            spoken_language,
            status,
            tagline,
            title,
            tmdb_id,
            vote_average,
            vote_count
        )
        VALUES (
            :budget,
            :cast_one,
            :cast_two,
            :cast_three,
            :color,
            :content,
            :director,
            :genre_one,
            :genre_two,
            :genre_three,
            :imdb_id,
            :notes,
            :original_language,
            :original_title,
            :overview,
            :popularity,
            :production_country,
            :release_date,
            :revenue,
            :runtime,
            :spoken_language,
            :status,
            :tagline,
            :title,
            :tmdb_id,
            :vote_average,
            :vote_count
        )
        """
    )
    db1_selection = select_all_data(con1_name)
    while db1_selection.next():
        for column_index in range(db1_selection.record().count()):
            if not db1_selection.record().fieldName(column_index) == "id":
                db2_insertion.addBindValue(db1_selection.value(column_index))
        db2_insertion.exec()
    db1_selection.finish()
    db2_insertion.finish()


def create_empty_table(con_name: str) -> None:
    """Create a table in a QSQLite database for storing film info.

    Parameters
    ----------
    con_name
        The name of a QSQLite database connection.

    Returns
    -------
    None

    """
    con = QSqlDatabase.database(con_name)
    query = QSqlQuery(con)
    query.exec(
        """
        CREATE TABLE films (
            budget INTEGER,
            cast_one TEXT,
            cast_two TEXT,
            cast_three TEXT,
            color TEXT,
            content TEXT,
            director TEXT,
            genre_one TEXT,
            genre_two TEXT,
            genre_three TEXT,
            id INTEGER PRIMARY KEY AUTOINCREMENT UNIQUE NOT NULL,
            imdb_id TEXT,
            notes TEXT,
            original_language TEXT,
            original_title TEXT,
            overview TEXT,
            popularity REAL,
            production_country TEXT,
            release_date TEXT,
            revenue INTEGER,
            runtime INTEGER,
            spoken_language TEXT,
            status TEXT,
            tagline TEXT,
            title TEXT,
            tmdb_id INTEGER,
            vote_average REAL,
            vote_count INTEGER
        )
        """
    )
    query.finish()


def open_db(con_name: str, db_name: str) -> bool:
    """Create a QSQLite database connection.

    Parameters
    ----------
    con_name
        The name of the QSQLite database connection to create.
    db_name
        The name (path) of the database to open.
        If ':memory:' is received, the database will be opened in-
        memory. If the name (path) provided does not exist, a disk-based
        database will be created in that location.

    Returns
    -------
    bool
        True if the connection is successfully opened, False if not.

    """
    con = QSqlDatabase.addDatabase("QSQLITE", con_name)
    con.setDatabaseName(db_name)
    if not con.open():
        MessageDialog(
            icon=QMessageBox.Icon.Warning,
            title="database error",
            text="An error occurred while opening the database.",
            info="The application will launch without an open database.",
            details=f"Error report:\n{con.lastError().databaseText()}",
            buttons=QMessageBox.Ok,
        ).exec()
        return False
    return True


def select_all_data(con_name: str) -> QSqlQuery:
    """Select all of the film information stored in a QSQLite database.

    Parameters
    ----------
    con_name
        The name of a QSQLite database connection.

    Returns
    -------
    :obj:`QSqlQuery`
        A query containing all of the information in the database.

    """
    con = QSqlDatabase.database(con_name)
    query = QSqlQuery(con)
    query.exec(
        """
        SELECT
            budget,
            cast_one,
            cast_two,
            cast_three,
            color,
            content,
            director,
            genre_one,
            genre_two,
            genre_three,
            id,
            imdb_id,
            notes,
            original_language,
            original_title,
            overview,
            popularity,
            production_country,
            release_date,
            revenue,
            runtime,
            spoken_language,
            status,
            tagline,
            title,
            tmdb_id,
            vote_average,
            vote_count
        FROM films
        ORDER BY id ASC
        """
    )
    return query


def main(argv: Optional[List[str]] = None) -> None:
    """Run the application's main graphic user interface.

    Parameters
    ----------
    argv
        An optional list of arguments to pass to QApplication.

    Returns
    -------
    None

    """
    app = QApplication([] if argv is None else argv)
    window = Interface()
    window.show()
    app.exec()


if __name__ == "__main__":
    main()



"""Initialize logging for later configuration.

Copyright (C) 2021 emerac

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
import logging


logger = logging.getLogger(__name__)
formatter = logging.Formatter(
    "%(relativeCreated)6d %(levelname)-8s %(message)s"
)
stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)
null_handler = logging.NullHandler()



"""Search for and parse film metadata from an external source.

Copyright (C) 2021 emerac

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
from typing import Any, Dict, List, Optional, cast

import requests


def get_cast(data: Dict[str, Any]) -> List[Optional[str]]:
    """Retrieve the top three cast members' names.

    Parameters
    ----------
    data
        Film credits obtained from TMDb.

    Returns
    -------
    list of str, optional
        The top three cast members' names, if they exist in the data.

    """
    top_three: List[Optional[str]] = [None for i in range(3)]
    for index, cast_member in enumerate(data["credits"]["cast"]):
        if index > 2:
            break
        top_three[index] = cast_member.get("name")
    return top_three


def get_director(data: Dict[str, Any]) -> Optional[Any]:
    """Retrieve the director's name.

    Parameters
    ----------
    data
        Film credits obtained from TMDb.

    Returns
    -------
    str, optional
        The director's name, if it exists in the data.

    """
    for member in data["credits"]["crew"]:
        if member["job"] == "Director":
            return member["name"]
    return None


def get_film_info(key: str, film_id: int) -> Dict[str, Any]:
    """Retrieve film details and credits.

    Parameters
    ----------
    key
        A valid TMDb API key.
    film_id
        The TMDb ID for a specific film.

    Returns
    -------
    dict
        JSON film details and credits for a specific film.

    Raises
    ------
    requests.exceptions.HTTPError
        If a bad request (4XX client or 5XX server error) was made
    requests.exceptions.Timeout
        If the server has not responded in N seconds

    """
    url = "https://api.themoviedb.org/3/movie/" + str(film_id)
    query_string = {
        "api_key": key,
        "append_to_response": "credits",
    }
    r = requests.get(url, params=query_string, timeout=3)
    r.raise_for_status()
    return cast(Dict[str, Any], r.json())


def get_genres(data: Dict[str, Any]) -> List[Optional[str]]:
    """Retrieve the top three genres.

    Parameters
    ----------
    data
        Film details obtained from TMDb.

    Returns
    -------
    list of str, optional
        The top three genres' names, if they exist in the data.

    """
    top_three: List[Optional[str]] = [None for i in range(3)]
    for index, genre in enumerate(data["genres"]):
        if index > 2:
            break
        top_three[index] = genre.get("name")
    return top_three


def get_search_results(
    key: str,
    query: str,
    year: str,
    page: int,
) -> Dict[str, Any]:
    """Retrieve search results from TMDb that match a query.

    Parameters
    ----------
    key
        A valid TMDb API key.
    query
        Text that may match against a film's title, description, etc.
    year
        Limit the results to films released in a particular year. If
        `year` is an empty string, all years will be considered.
    page
        Display a particular page of search results.

    Returns
    -------
    dict
        JSON search results for a certain page, with approximately 20
        results per page.

    Raises
    ------
    requests.exceptions.HTTPError
        If a bad request (4XX client or 5XX server error) was made
    requests.exceptions.Timeout
        If the server has not responded in N seconds

    """
    url = "https://api.themoviedb.org/3/search/movie"
    query_string = {
        "api_key": key,
        "query": query,
        "year": year,
        "page": str(page),
    }
    r = requests.get(url, params=query_string, timeout=3)
    r.raise_for_status()
    return cast(Dict[str, Any], r.json())



"""A window for searching for, choosing, and adding film metadata.

Copyright (C) 2021 emerac

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
from typing import Any, Dict, Optional

from PyQt5.QtGui import (
    QCloseEvent,
    QIntValidator,
    QShowEvent,
)
from PyQt5.QtWidgets import (
    QApplication,
    QCheckBox,
    QFormLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMainWindow,
    QMessageBox,
    QProgressBar,
    QPushButton,
    QSizePolicy,
    QTreeWidget,
    QTreeWidgetItem,
    QVBoxLayout,
    QWidget,
)
import requests

from filmbuff import extdata


class EmptyErrorDialog(QMessageBox):
    """A dialog for displaying errors resulting from empty fields."""

    def __init__(self, field: str, parent: Optional[QWidget] = None) -> None:
        """Construct an EmptyErrorDialog instance.

        Parameters
        ----------
        field
            The field that is missing contents.
        parent
            The EmptyErrorDialog's parent widget.

        Returns
        -------
        None

        """
        super().__init__(parent)
        self.setWindowTitle("Search error")
        self.setText(f"The {field} field is required.")
        self.setInformativeText(f"Please enter a {field} and try again.")
        self.setIcon(QMessageBox.Warning)
        self.setStandardButtons(QMessageBox.Ok)


class MetadataWindow(QWidget):
    """A window for retrieving metadata from an external source."""

    def __init__(self, main_win: QMainWindow) -> None:
        """Construct a MetadataWindow instance.

        Parameters
        ----------
        main_win
            The MetadataWindow's parent window.

        Returns
        -------
        None

        """
        super().__init__()
        self.main_win = main_win

        self.setWindowTitle("Retrieve metadata - filmbuff")
        self.resize(500, 600)

        query_label = QLabel("Query")
        self.query_field = QLineEdit()
        self.query_field.textChanged.connect(self.field_changed)
        year_label = QLabel("Year (optional)")
        self.year_field = QLineEdit()
        self.year_field.setValidator(QIntValidator(1, 3000))
        self.year_field.textChanged.connect(self.field_changed)
        key_label = QLabel("TMDB API key")
        self.key_field = QLineEdit()
        self.key_field.setEchoMode(QLineEdit.Password)
        self.key_field.textChanged.connect(self.field_changed)
        remember_label = QLabel("Remember key for session")
        self.remember_box = QCheckBox()
        self.remember_box.setSizePolicy(
            QSizePolicy.Maximum,
            QSizePolicy.Preferred,
        )
        self.search_button = QPushButton("Search")
        self.search_button.setAutoDefault(True)
        self.search_button.pressed.connect(self.search_pressed)
        self.progressbar = QProgressBar()
        self.progressbar.hide()
        self.tree = QTreeWidget()
        self.tree.setColumnCount(5)
        self.tree.setHeaderLabels(["ID", "Year", "Title", "Lead", "Director"])
        self.tree.hideColumn(0)
        self.tree.setIndentation(0)
        self.tree.itemDoubleClicked.connect(self.add_item)
        self.tree.itemSelectionChanged.connect(self.toggle_add_button)
        self.results_label = QLabel("No results currently displayed.")
        self.load_button = QPushButton("Load more")
        self.load_button.setAutoDefault(True)
        self.load_button.setDisabled(True)
        self.load_button.pressed.connect(self.load_pressed)
        self.add_button = QPushButton("Add data")
        self.add_button.setAutoDefault(True)
        self.add_button.setDisabled(True)
        self.add_button.pressed.connect(self.add_item)
        self.search_spacer = QWidget()
        self.search_spacer.setSizePolicy(
            QSizePolicy.Expanding,
            QSizePolicy.Preferred,
        )
        button_spacer = QWidget()
        button_spacer.setSizePolicy(
            QSizePolicy.Expanding,
            QSizePolicy.Preferred,
        )

        form_layout = QFormLayout()
        form_layout.setRowWrapPolicy(QFormLayout.WrapAllRows)
        form_layout.addRow(query_label, self.query_field)
        form_layout.addRow(year_label, self.year_field)
        form_layout.addRow(key_label, self.key_field)
        remember_layout = QHBoxLayout()
        remember_layout.addWidget(self.remember_box)
        remember_layout.addWidget(remember_label)
        search_layout = QHBoxLayout()
        search_layout.addWidget(self.search_button)
        search_layout.addWidget(self.search_spacer)
        search_layout.addWidget(self.progressbar)
        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(self.load_button)
        buttons_layout.addWidget(button_spacer)
        buttons_layout.addWidget(self.add_button)
        main_layout = QVBoxLayout()
        main_layout.addLayout(form_layout)
        main_layout.addLayout(remember_layout)
        main_layout.addLayout(search_layout)
        main_layout.addWidget(self.tree)
        main_layout.addWidget(self.results_label)
        main_layout.addLayout(buttons_layout)

        self.setLayout(main_layout)

        self.results: Optional[Dict[str, Any]] = {}
        self.query = ""
        self.year = ""
        self.key = ""
        self.next_page = 1

    def closeEvent(self, event: QCloseEvent) -> None:
        """Handle QCloseEvents.

        This method waits for QCloseEvents to occur. Given a
        QCloseEvent, it will ensure that all fields are cleared and
        all attributes reset. The key field will not be reset if its
        corresponding checkbox is checked.

        Parameters
        ----------
        event
            Information about the close event that occurred.

        Returns
        -------
        None

        Notes
        -----
        This is an overridden baseclass (QWidget) method.

        This method handles QCloseEvents, which are automatically
        produced in response to widget close signals. When a QCloseEvent
        occurs, the QCloseEvent is automatically forwarded to this
        method. QCloseEvents can be manufactured and passed to this
        method as well.

        """
        super().closeEvent(event)
        self.results = {}
        self.query = ""
        self.query_field.setText(self.query)
        self.year = ""
        self.year_field.setText(self.year)
        if not self.remember_box.isChecked():
            self.key = ""
            self.key_field.setText(self.key)
        self.results_label.setText("No results currently displayed.")
        self.tree.clear()
        self.next_page = 1

    def showEvent(self, event: QShowEvent) -> None:
        """Handle QShowEvents.

        This method waits for QShowEvents to occur. When a QShowEvent is
        received, the window will be made to be centered over the parent
        window.

        Parameters
        ----------
        event
            Information about the show event that occurred.

        Returns
        -------
        None

        Notes
        -----
        This is an overridden baseclass (QWidget) method.

        This method handles QShowEvents, which are automatically
        produced in response to widget show signals. When a QShowEvent
        occurs, the QShowEvent is automatically forwarded to this
        method. QShowEvents can be manufactured and passed to this
        method as well.

        """
        super().showEvent(event)
        self.center_window()
        self.query_field.setFocus()

    def add_item(self) -> None:
        """Add a filled-in record to the parent window's model.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Notes
        -----
        Because the tree widget only contains a portion of all the
        information available, choosing to add a particular tree widget
        item will make a request for all of that item's information.

        """
        selection = self.tree.selectedItems()
        selection_id = int(selection[0].text(0))

        info = extdata.get_film_info(self.key, selection_id)
        cast = extdata.get_cast(info)
        director = extdata.get_director(info)
        genres = extdata.get_genres(info)
        record_data = [
            info["budget"],
            cast[0],
            cast[1],
            cast[2],
            None,
            None,
            director,
            genres[0],
            genres[1],
            genres[2],
            None,
            info["imdb_id"],
            None,
            info["original_language"],
            info["original_title"],
            info["overview"],
            info["popularity"],
            info["production_countries"][0]["iso_3166_1"],
            info["release_date"],
            info["revenue"],
            info["runtime"],
            info["spoken_languages"][0]["iso_639_1"],
            info["status"],
            info["tagline"],
            info["title"],
            info["id"],
            info["vote_average"],
            info["vote_count"],
        ]
        new_record = self.main_win.model.record(-1)
        for i in range(new_record.count()):
            if new_record.fieldName(i) == "id":
                continue
            new_record.setValue(i, record_data[i])
        self.main_win.insert_rows(False, record=new_record)

    def center_window(self) -> None:
        """Center the window over its parent window.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        frame_width = self.main_win.frameGeometry().width()
        win_width = self.main_win.width()
        header_width = frame_width - win_width

        frame_height = self.main_win.frameGeometry().height()
        win_height = self.main_win.height()
        header_height = frame_height - win_height

        center_x = self.main_win.x() + header_width + 0.5 * win_width
        center_y = self.main_win.y() + header_height + 0.5 * win_height
        win_x = center_x - 0.5 * self.width()
        win_y = center_y - 0.5 * self.height()
        self.move(int(win_x), int(win_y))

    def field_changed(self) -> None:
        """Toggle the state of the load button based on field contents.

        When any of the fields are changed, disable the load button to
        prevent results from a different query from being returned.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        if (
            not self.query_field.text() == self.query
            or not self.year_field.text() == self.year
            or not self.key_field.text() == self.key
        ):
            self.load_button.setDisabled(True)
        else:
            self.toggle_load_button()

    def load_pressed(self) -> None:
        """Perform a search when the load button is pressed.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Notes
        -----
        This method directly calls `perform_search`, which increments
        the page number after each call. When this method is called, it
        is assumed that the page number has been incremented and that
        the page count is greater than one.

        """
        self.perform_search()

    def perform_search(self) -> None:
        """Perform a search when the search button is pressed.

        This method uses the information provided in the fields to
        obtain search results from an external source. The results are
        then displayed in the tree widget.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        self.results = self.retrieve_results()
        if self.results is None:
            return

        self.search_spacer.hide()
        self.progressbar.show()
        # Normally, events are processed only after a function ends, but
        # the show/hide calls above need to be performed right now.
        QApplication.processEvents()
        self.progressbar.setMinimum(0)
        self.progressbar.setMaximum(len(self.results["results"]))

        for i, r in enumerate(self.results["results"]):
            r_info = extdata.get_film_info(self.key, r["id"])
            r_id = str(r_info["id"])
            r_year = r_info["release_date"][:4]
            r_title = r_info["title"]
            r_lead = extdata.get_cast(r_info)[0]
            r_director = extdata.get_director(r_info)

            raw_data = [r_id, r_year, r_title, r_lead, r_director]
            data = [x if x is not None else "Unavailable" for x in raw_data]

            item = QTreeWidgetItem(data)
            self.tree.addTopLevelItem(item)
            self.progressbar.setValue(i)

        self.results_label.setText(
            f"Displaying {self.tree.topLevelItemCount()} of "
            f"{self.results['total_results']} results."
        )

        # Subsequent searches should obtain subsequent pages.
        self.next_page = self.next_page + 1
        self.toggle_load_button()

        self.search_spacer.show()
        self.progressbar.reset()
        self.progressbar.hide()
        self.resize_columns()

    def resize_columns(self) -> None:
        """Ensure columns do not get pushed out of the viewport.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        for i in range(self.tree.columnCount()):
            self.tree.resizeColumnToContents(i)
            if self.tree.columnWidth(i) > 150:
                self.tree.setColumnWidth(i, 150)
            if not i == self.tree.columnCount() - 1:
                new_width = self.tree.columnWidth(i) + 16
                self.tree.setColumnWidth(i, new_width)

    def retrieve_results(self) -> Optional[Dict[str, Any]]:
        """Obtain search results from an external source.

        Parameters
        ----------
        None

        Returns
        -------
        optional dict of str and Any
            Search results in json form.

        Raises
        ------
        requests.exceptions.HTTPError
            If a 401 (unauthorized) or 404 (connection) error occurred.

        """
        try:
            results = extdata.get_search_results(
                self.key,
                self.query,
                page=self.next_page,
                year=self.year,
            )
        except requests.exceptions.HTTPError as e:
            UnauthorizedErrorDialog(e, self).exec()
            return None
        return results

    def search_pressed(self) -> None:
        """Perform a search when the search button is pressed.

        This method clears the tree widget of any contents to
        accommodate new search results.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        self.query = self.query_field.text()
        if self.query == "":
            EmptyErrorDialog("query", self).exec()
            return
        self.year = self.year_field.text()
        self.key = self.key_field.text()
        if self.key == "":
            EmptyErrorDialog("key", self).exec()
            return

        self.tree.clear()
        self.next_page = 1
        self.perform_search()

    def toggle_add_button(self) -> None:
        """Change the add button's state based on the tree selection.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        if self.tree.selectedItems():
            self.add_button.setDisabled(False)
        else:
            self.add_button.setDisabled(True)

    def toggle_load_button(self) -> None:
        """Change the load button's state based on the search results.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        if self.results:
            if self.next_page <= self.results["total_pages"]:
                self.load_button.setDisabled(False)
            else:
                self.load_button.setDisabled(True)


class UnauthorizedErrorDialog(QMessageBox):
    """A dialog for displaying errors resulting from bad requests."""

    def __init__(
        self,
        error: requests.exceptions.HTTPError,
        parent: Optional[QWidget] = None,
    ) -> None:
        """Construct an UnauthorizedErrorDialog instance.

        Parameters
        ----------
        error
            The HTTPError object provided by `requests.get`.
        parent
            The EmptyErrorDialog's parent widget.

        Returns
        -------
        None

        """
        super().__init__(parent)
        self.setWindowTitle("Search error")
        self.setText(
            f"Error {error.response.status_code} - {error.response.reason}",
        )
        self.setInformativeText(
            "The supplied key was rejected. It may contain a typo "
            "or it may no longer be valid."
        )
        self.setDetailedText(
            "Status message returned:\n"
            f"{error.response.json()['status_message']}\n"
        )
        self.setIcon(QMessageBox.Warning)
        self.setStandardButtons(QMessageBox.Ok)
